use std::collections::HashMap;

use super::{
    expr::{Expr, ExprOp},
    opcode::OpCode,
    parser::parse_query,
    runtime::execute,
    value::Value,
    Addr, Error, Register,
};
use crate::Transaction;

#[derive(Debug, Clone)]
pub struct Program {
    /// Debug op-codes, with the unresolved labels.
    dbg_opcodes: Vec<OpCode>,
    /// The list of opcodes that make up the program
    opcodes: Vec<OpCode>,
    /// If the program has some boilerplate that can be skipped. This is non-zero when the initial
    /// program has been executed at compile time, and the register has been populated to
    /// `initial_register`. Everytime a new Runtime is created, instead of executing the program,
    /// it can safely start-off a different Address, and clone the initial_register.
    start_at: Addr,
    /// The state of the register
    initial_register: Vec<Value>,
}

pub struct Compiler<'a> {
    expr: &'a Expr,
    current_register: Register,
    labels: Addr,
}

impl<'a> Compiler<'a> {
    pub fn new(expr: &'a Expr) -> Self {
        Self {
            expr,
            current_register: 0,
            labels: 0,
        }
    }

    fn next_label(&mut self) -> Addr {
        self.labels += 1;
        self.labels
    }

    fn next_register(&mut self) -> usize {
        self.current_register += 1;
        self.current_register
    }

    fn compile_expr(&mut self, expr: &'a Expr, exit_label: Addr) -> (Vec<OpCode>, usize) {
        let return_value = self.next_register();
        match expr {
            Expr::Variable(name) => (
                vec![OpCode::LOAD_EXTERNAL(return_value, name.into())],
                return_value,
            ),
            Expr::String(string) => (
                vec![OpCode::LOAD(return_value, Value::String(string.clone()))],
                return_value,
            ),
            Expr::Number(number) => (
                vec![OpCode::LOAD(return_value, Value::Number(*number))],
                return_value,
            ),
            Expr::Not(expr) => {
                let (mut expr, prev_return_value) = self.compile_expr(expr, exit_label);
                expr.push(OpCode::NOT(return_value, prev_return_value));
                (expr, return_value)
            }
            Expr::Op { op, left, right } => {
                let mut to_return = vec![];
                let current_exit_label = self.next_label();
                let (mut left, return_left) = self.compile_expr(left, current_exit_label);
                let (mut right, return_right) = self.compile_expr(right, current_exit_label);

                to_return.append(&mut left.clone());
                to_return.append(&mut right.clone());

                match op {
                    ExprOp::Eq => {
                        to_return.push(OpCode::EQ(return_value, return_left, return_right));
                    }
                    ExprOp::Add => {
                        to_return.push(OpCode::ADD(return_value, return_left, return_right))
                    }
                    ExprOp::Or => {
                        to_return = vec![OpCode::LOAD(return_value, Value::Bool(false))];
                        to_return.append(&mut left);
                        to_return.push(OpCode::MOV(return_value, return_left));
                        to_return.push(OpCode::JEQ(return_value, exit_label));
                        to_return.append(&mut right);
                        to_return.push(OpCode::OR(return_value, return_left, return_right))
                    }
                    ExprOp::And => {
                        to_return = vec![OpCode::LOAD(return_value, Value::Bool(false))];
                        to_return.append(&mut left);
                        to_return.push(OpCode::MOV(return_value, return_left));
                        to_return.push(OpCode::JNE(return_value, exit_label));
                        to_return.append(&mut right);
                        to_return.push(OpCode::AND(return_value, return_left, return_right))
                    }
                    ExprOp::Mul => {
                        to_return.push(OpCode::MUL(return_value, return_left, return_right))
                    }
                    ExprOp::Sub => {
                        to_return.push(OpCode::SUB(return_value, return_left, return_right))
                    }
                    x => panic!("unimplemented {:#?}", x),
                }

                to_return.push(OpCode::LABEL(current_exit_label));
                (to_return, return_value)
            }
        }
    }

    fn resolve_label_to_addr(opcodes: Vec<OpCode>) -> Result<Vec<OpCode>, Error> {
        let mut pos = 0;
        let used_labels = opcodes
            .iter()
            .map(|x| match x {
                OpCode::LABEL(label) => (x, pos),
                _ => {
                    pos = pos + 1;
                    (x, pos - 1)
                }
            })
            .filter_map(|(opcode, pos)| match opcode {
                OpCode::LABEL(id) => Some((*id, pos)),
                _ => None,
            })
            .collect::<HashMap<_, _>>();

        opcodes
            .into_iter()
            .filter(|opcode| !matches!(opcode, OpCode::LABEL(_)))
            .map(|opcode| {
                Ok(match opcode {
                    OpCode::JMP(label) => {
                        OpCode::JMP(*used_labels.get(&label).ok_or(Error::UnknownLabel(label))?)
                    }
                    OpCode::JEQ(register, label) => OpCode::JEQ(
                        register,
                        *used_labels.get(&label).ok_or(Error::UnknownLabel(label))?,
                    ),
                    OpCode::JNE(register, label) => OpCode::JNE(
                        register,
                        *used_labels.get(&label).ok_or(Error::UnknownLabel(label))?,
                    ),
                    x => x,
                })
            })
            .collect()
    }

    pub fn compile(mut self) -> Vec<OpCode> {
        let exit_point = self.next_label();
        let (mut opcodes, return_value) = self.compile_expr(self.expr, exit_point);
        opcodes.push(OpCode::LABEL(exit_point));
        opcodes.push(OpCode::HLT(return_value));
        opcodes
    }
}

impl Program {
    pub fn new(code: &str) -> Result<Self, Error> {
        let ast = parse_query(code)?;

        let opcodes = ast.where_clause.map_or_else(
            || vec![OpCode::LOAD(0, true.into()), OpCode::HLT(0)],
            |expr| Compiler::new(&expr).compile(),
        );

        Ok(Self {
            dbg_opcodes: opcodes.clone(),
            opcodes: Compiler::resolve_label_to_addr(opcodes)?,
            start_at: 0,
            initial_register: vec![],
        })
    }

    /// Returns a human readable version of the compiled program (generated op-codes)
    pub fn debug(&self) -> String {
        self.dbg_opcodes
            .iter()
            .map(|x| match x {
                OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                x => format!("\t{}", x.to_string()),
            })
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn dump(&self) -> String {
        self.opcodes
            .iter()
            .enumerate()
            .map(|(pos, opcode)| format!("{}: {}", pos, opcode.to_string()))
            .collect::<Vec<_>>()
            .join("\n")
    }

    /// Returns the opcodes of the program expression and the usize of the register where the result
    /// Creates a program tailored for a transaction
    pub fn execute(&self, transaction: Option<&Transaction>) -> Result<Value, Error> {
        execute(
            transaction,
            &self.opcodes,
            self.initial_register.clone(),
            self.start_at,
        )
    }
}

#[cfg(test)]
mod test {
    use super::Program;
    use crate::filter_expr::{opcode::OpCode, value::Value};

    #[test]
    fn parse() {
        let x = Program::new(
            r#"
            WHERE
                $foo = 3
                AND $bar = "bar"
                AND ($a = $b OR $b = $c)
                AND $foo = 1 + 2 + 3 +  ((1+ 2-3*4-5) * $bar.tx.lol)
            ORDER BY $bar DESC
        "#,
        )
        .unwrap();
        panic!("{}\n\n{}", x.debug(), x.dump());
    }

    #[test]
    fn simple_program() {
        let program = Program {
            dbg_opcodes: vec![],
            opcodes: vec![
                OpCode::LOAD(1, 12.into()),
                OpCode::LOAD(2, 13.into()),
                OpCode::ADD(3, 1, 2),
                OpCode::ADD(4, 0, 3),
                OpCode::HLT(4),
            ],
            start_at: 0,
            initial_register: vec![15.into()],
        };
        let x = program.execute(None).expect("valid execution");
        assert_eq!(x, 40.into());
    }
}
